unit Main;

interface
uses Classes, SysUtils, Struct;

type
  TEntryType = (seLanguage, sePermission, seType, seComponent, seTask, seDir,
    seFile, seFileLocation, seIcon, seIni, seRegistry, seInstallDelete,
    seUninstallDelete, seRun, seUninstallRun, seCustomMessage);
  TMakeDirFlags = set of (mdNoUninstall, mdAlwaysUninstall, mdDeleteAfterInstall,
    mdNotifyChange);

  TCommandAction=(caInstallInfo, {caListFiles,} caVerboseList, caExtractFiles, caWriteCode);

  EMessage = class(Exception); // Exception raised by our code, The Message field (string)
                               // holds all necessary information, No need to display
                               // the exception class, ErrorMessage field, Win32 last error,
                               // or (try to) track down error address.

  EFatalError = class(Exception); // Halt program. Only handled by the top-level handler,
                                  // not by local handlers. Typically raised when an error
                                  // has been caught and processed but the execution can not
                                  // continue.
const
//  SSetupFileCorrupt = 'The setup files are corrupted. Please obtain a new ' +
//    'copy of the program.';
  SelfReadMode = fmOpenRead or fmShareDenyWrite;

{  EntryStrings: array[TEntryType] of Integer = (SetupLanguageEntryStrings,
    SetupPermissionEntryStrings, SetupTypeEntryStrings,
    SetupComponentEntryStrings, SetupTaskEntryStrings, SetupDirEntryStrings,
    SetupFileEntryStrings, SetupFileLocationEntryStrings,
    SetupIconEntryStrings, SetupIniEntryStrings, SetupRegistryEntryStrings,
    SetupDeleteEntryStrings, SetupDeleteEntryStrings, SetupRunEntryStrings,
    SetupRunEntryStrings);}

var
  SetupLdrOffset0: integer;
  SetupLdrOffset1: integer;
  SetupLdrMode: Boolean = true;
  SourceDir: String;
  SetupLdrOriginalFilename: String;
  Debugging: Boolean = false;
  Entries: array[TEntryType] of TList;
  SetupHeader: TSetupHeader;
  FileVersion:string; // based on TestID
  TimeStampsInUTC:boolean;

  StdOutputHandle: THandle;

  CommandAction: TCommandAction;
  QuietExtract:boolean=false;
  ExtractEmbedded:boolean=false;
  StripPaths:boolean=false;
  InteractiveMode:boolean=true;
  OutDir:string='';
  BaseDirToStrip:string='';
  AttemptUnpackUnknown:boolean=true;
  ExtractAllCopies:boolean=false;
  WizardImage, WizardSmallImage, DecompDll:string;
  AutoYes: boolean = false;
  ExtractTestOnly: boolean = false;

procedure InternalError(const Id: String);
function TestPassword(const Password: String): Boolean;
function GetSliceName(ASlice:integer):string;
function FixPasswordEncoding(const Password: String) : String;

function StrToOem(const s:string):string;
function OemToStr(const s:string):string;
procedure write(s:string);
procedure writeln(const s:string);

function AddFakeFile(FileName,FileContents:string; RenameNow:boolean=false):integer;

function FileContents(Filename:string):string;

implementation

uses MD5, SHA1, PathFunc, MyTypes, Windows;

procedure InternalError(const Id: String);
begin
  raise Exception.Create(Id);
end;

function TestPassword_MD5(const Password: String): Boolean;
var
  Context: TMD5Context;
  Hash: TMD5Digest;
begin
  MD5Init(Context);
  MD5Update(Context, PAnsiChar('PasswordCheckHash')^, Length('PasswordCheckHash'));
  MD5Update(Context, SetupHeader.PasswordSalt, SizeOf(SetupHeader.PasswordSalt));
  MD5Update(Context, Pointer(Password)^, Length(Password));
  Hash := MD5Final(Context);
  Result := MD5DigestsEqual(Hash, SetupHeader.PasswordHash_MD5);
end;

function TestPassword_SHA1(const Password: String): Boolean;
var
  Context: TSHA1Context;
  Hash: TSHA1Digest;
begin
  SHA1Init(Context);
  SHA1Update(Context, PAnsiChar('PasswordCheckHash')^, Length('PasswordCheckHash'));
  SHA1Update(Context, SetupHeader.PasswordSalt, SizeOf(SetupHeader.PasswordSalt));
  SHA1Update(Context, Pointer(Password)^, Length(Password)*SizeOf(Password[1]));
  Hash := SHA1Final(Context);
  Result := SHA1DigestsEqual(Hash, SetupHeader.PasswordHash_SHA1);
end;

function TestPassword(const Password: String): Boolean;
begin
	if (Ver >= 5309) then
		Result := TestPassword_SHA1(Password)
	else
		Result := TestPassword_MD5(Password);
end;

function GetSliceName(ASlice:integer):string;
var
  Major, Minor: Integer;
  Prefix:string;
begin
  if Ver>=4107 then Prefix := PathChangeExt(PathExtractName(SetupLdrOriginalFilename), '')
  else Prefix:=SetupHeader.BaseFilename;
  Major := ASlice div SetupHeader.SlicesPerDisk + 1;
  Minor := ASlice mod SetupHeader.SlicesPerDisk;
  if SetupHeader.SlicesPerDisk = 1 then
    Result := Format('%s-%d.bin', [Prefix, Major])
  else
    Result := Format('%s-%d%s.bin', [Prefix, Major, Chr(Ord('a') + Minor)]);
end;

function StrToOem(const s:string):string;
begin
  Result:=s; UniqueString(Result);
  CharToOemBuff(PChar(Result),PChar(Result),length(Result));
end;

function OemToStr(const s:string):string;
begin
  Result:=s; UniqueString(Result);
  OemToCharBuff(PChar(Result),PChar(Result),length(Result));
end;

// Delphi RTL does not check GetLastError() after getting GetFileType(hFile)==FILE_TYPE_UNKNOWN
// and incorrectly decides that an error occurred.
// This condition was observed when standard output was redirected by the parent process.
// Here is a workaround. Note that write/writeln now take exactly one argument of type string.
// It's also a good place to convert the characters into the OEM codepage.
// In case standard output is not a disk file, writes are split into 1K chunks to avoid
// out of memory errors.
procedure write(s:string);
var
  WriteOffset,BytesRemaining,BytesToWrite,BytesWritten:dword;
begin
  s:=StrToOem(s);
  WriteOffset:=0; BytesRemaining:=length(s);
  while BytesRemaining>0 do begin
    if BytesRemaining>1024 then BytesToWrite:=1024 else BytesToWrite:=BytesRemaining;
    WriteFile(StdOutputHandle,s[WriteOffset+1],BytesToWrite,BytesWritten,nil);
{    if BytesWritten<>BytesToWrite then begin
      system.writeln('WriteFile failed. Requested=',BytestoWrite,' written=',BytesWritten,' error=',GetLastError);
      halt;
    end;}
    Inc(WriteOffset,BytesToWrite); Dec(BytesRemaining,BytesToWrite);
  end;
end;

procedure writeln(const s:string);
begin
  write(s+#13#10);
end;

function AddFakeFile(FileName,FileContents:string; RenameNow:boolean=false):integer;
var
  pFileEntry:PSetupFileEntry;
  pFileLocationEntry:PSetupFileLocationEntry;
  SystemTime:TSystemTime;
  i:integer;
begin
  pFileLocationEntry:=AllocMem(sizeof(TSetupFileLocationEntry));
  with pFileLocationEntry^ do begin
    Contents:=FileContents;
    Int64(OriginalSize):=length(FileContents);
    DateTimeToSystemTime(Now,SystemTime);
    SystemTimeToFileTime(SystemTime,TimeStamp);
    PrimaryFileEntry:=-1;
  end;
  i:=Entries[seFileLocation].Add(pFileLocationEntry);
  pFileEntry:=AllocMem(sizeof(TSetupFileEntry));
  with pFileEntry^ do begin
    FileType:=ftFakeFile;
    DestName:=FileName; // RenameFiles() will split this
    LocationEntry:=i;
    if RenameNow then begin
      SourceFilename:=FileName; DestName:=''; DestDir:='';
    end;
  end;

  Result:=Entries[seFile].Add(pFileEntry);
end;

function FileContents(Filename:string):string;
var f:File;
begin
  AssignFile(f, Filename); Result:='';
  Reset(f,1);
  SetLength(Result, FileSize(f));
  BlockRead(f, Result[1], FileSize(f));
  CloseFile(f);
end;

function FixPasswordEncoding(const Password: String) : String;
var
  tmpWS: WideString;
  nWideBytesCount: integer;
begin
  Result := Password;
  if (VerIsUnicode) then
  begin
    tmpWS := Password;

    nWideBytesCount := Length(Result) * SizeOf(WideChar);
    SetLength(Result, nWideBytesCount);
    CopyMemory(@Result[1], PAnsiChar(@tmpWS[1]), nWideBytesCount);
  end;
end;

end.